import ast

_IW_BULK_UPDATE_WITHOUT_BATCH_SIZE_ERROR_MSG = "IW06 bulk update without batch_size"
_BULK_UPDATE_METHOD_CALLS = ["bulk_update", "bulk_create"]


class BulkUpdateFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        # bulk_create / bulk_update
        # Model.objects.bulk_update()
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Attribute)
            and isinstance(node.func.value.value, ast.Name)
            and node.func.value.attr == "objects"
            and node.func.attr in _BULK_UPDATE_METHOD_CALLS
        ):
            if not any([keywork.arg == "batch_size" for keywork in node.keywords]):
                self.issues.append(
                    (node.func.lineno, node.func.col_offset, _IW_BULK_UPDATE_WITHOUT_BATCH_SIZE_ERROR_MSG)
                )

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_SHARED_TASK_MSG = "IW07 use of celery.shared_task, use instawork.decorators.shared_task"


class SharedTaskFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        # from celery import shared_task
        if node.module == "celery":
            imports_shared_task = any(map(lambda x: x.name == "shared_task", node.names))
            if imports_shared_task:
                self.issues.append((node.lineno, node.col_offset, _IW_ERROR_SHARED_TASK_MSG))

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_FACTORY_TIMEZONE_NOW_MSG = "IW15 use of timezone.now in factory, call from LazyAttribute or LazyFunction"


class FactoryTimezoneNowFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        self._is_factory = False
        self._is_lazy_fn_or_attr = False

        super().__init__()

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        is_factory = node.name.endswith("Factory")
        if is_factory:
            self._is_factory = True

        self.generic_visit(node)
        if is_factory:
            self._is_factory = False
        return node

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        # from django.utils import timezone
        # class TestFactory(factory.django.DjangoModelFactory):
        #    starts_at = timezone.now()
        if (self._is_factory and isinstance(node.value, ast.Call)
                and isinstance(node.value.func, ast.Attribute)
                and isinstance(node.value.func.value, ast.Name)
                and node.value.func.value.id == "timezone"
                and node.value.func.attr == "now"):
            self.issues.append(
                (node.value.func.value.lineno, node.value.func.value.col_offset, _IW_ERROR_FACTORY_TIMEZONE_NOW_MSG)
            )

        # from django.utils.timezone import now
        # class TestFactory(factory.django.DjangoModelFactory):
        #    starts_at = now()
        if (
                self._is_factory
                and isinstance(node.value, ast.Call)
                and isinstance(node.value.func, ast.Name)
                and node.value.func.id == "now"
        ):
            self.issues.append(
                (node.value.func.lineno, node.value.func.col_offset, _IW_ERROR_FACTORY_TIMEZONE_NOW_MSG)
            )

        self.generic_visit(node)

        return node



import ast

_IW_ERROR_FOREIGN_KEY_CONSTRAINT = "IW09 ForeignKey without db_constraint=False"
_IW_ERROR_FOREIGN_KEY_ON_DELETE = "IW10 ForeignKey with on_delete=DO_NOTHING"

FIELD_TYPES = {"ForeignKey", "OneToOneField"}


class ForeignKeyCheckFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        is_foreign_key = (isinstance(node.func, ast.Attribute) and node.func.attr in FIELD_TYPES) or (
            isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id in FIELD_TYPES
        )

        if is_foreign_key:
            if not any([keyword.arg == "db_constraint" and keyword.value.value is False for keyword in node.keywords]):
                self.issues.append(
                    (node.func.lineno, node.func.col_offset, _IW_ERROR_FOREIGN_KEY_CONSTRAINT)
                )

            for keyword in node.keywords:
                if not keyword.arg == "on_delete":
                    continue
                if isinstance(keyword.value, ast.Name) and keyword.value.id == "DO_NOTHING":
                    self.issues.append(
                        (node.func.lineno, node.func.col_offset, _IW_ERROR_FOREIGN_KEY_ON_DELETE)
                    )
                elif isinstance(keyword.value, ast.Attribute) and keyword.value.attr == "DO_NOTHING":
                    self.issues.append(
                        (node.func.lineno, node.func.col_offset, _IW_ERROR_FOREIGN_KEY_ON_DELETE)
                    )

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_DATETIME_REPLACE_TZINFO_MSG = "IW05 use of datetime.replace(tzinfo=XXX)"


class DatetimeReplaceTzinfoFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        # dt.replace(tzinfo=X)
        if isinstance(node.func, ast.Attribute) and node.func.attr == "replace":
            if any([keyword.arg == "tzinfo" for keyword in node.keywords]):
                self.issues.append(
                    (node.func.value.lineno, node.func.value.col_offset, _IW_ERROR_DATETIME_REPLACE_TZINFO_MSG)
                )

        self.generic_visit(node)
        return node



import ast
from typing import Any

_IW_ERROR_SETUPCLASS_MSG = "IW16 do not use setUpClass and tearDownClass inside a test case"
_METHODS_TO_INSPECT = ["setUpClass", "tearDownClass"]


class PytestSetupAndTearDownFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []
        self.is_test_case = False
        self.is_method_to_inspect = False
        self.has_super_call = False

        super().__init__()

    def visit_ClassDef(self, node: ast.Attribute) -> ast.ClassDef:
        # Check if class is a TestCase as it should have setUpClass or tearDownClass
        self.is_test_case = any(isinstance(base, ast.Name) and base.id.find("TestCase") != -1 for base in node.bases)
        self.generic_visit(node)

        if self.is_test_case:
            self.is_test_case = False

        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        # Check if function is one of the methods to inspect and is in the test class
        self.has_setupclass_or_teardownclass = self.is_test_case and node.name in _METHODS_TO_INSPECT

        # after visiting nodes inside the function check if a super call was made
        if self.has_setupclass_or_teardownclass:
            self.issues.append((node.lineno, node.col_offset, _IW_ERROR_SETUPCLASS_MSG))

        return node



import ast

_IW_ERROR_DB_INDEX = "IW11 Use of db_index=True"
_IW_ERROR_UNIQUE = "IW12 Use of unique=True"

FIELD_TYPES = {
    "ForeignKey",
    "OneToOneField",
    "CharField",
    "BigIntegerField",
    "BinaryField",
    "BooleanField",
    "DateField",
    "DateTimeField",
    "DecimalField",
    "DurationField",
    "EmailField",
    "FileField",
    "FilePathField",
    "FloatField",
    "GenericIPAddressField",
    "ImageField",
    "IntegerField",
    "NullBooleanField",
    "PositiveIntegerField",
    "PositiveSmallIntegerField",
    "SlugField",
    "SmallIntegerField",
    "TextField",
    "TimeField",
    "URLField",
    "UUIDField",
}


class ConstraintCheckFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        is_foreign_key = (isinstance(node.func, ast.Attribute) and node.func.attr in FIELD_TYPES) or (
            isinstance(node, ast.Call) and isinstance(node.func, ast.Name) and node.func.id in FIELD_TYPES
        )

        if is_foreign_key:
            if any([keyword.arg == "db_index" and keyword.value.value is True for keyword in node.keywords]):
                self.issues.append(
                    (node.func.lineno, node.func.col_offset, _IW_ERROR_DB_INDEX)
                )

            if any([keyword.arg == "unique" and keyword.value.value is True for keyword in node.keywords]):
                self.issues.append(
                    (node.func.lineno, node.func.col_offset, _IW_ERROR_UNIQUE)
                )

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_INDEX_TOGETHER = "IW13 Use of index_together instead of Options.indexes"
_IW_ERROR_UNIQUE_TOGETHER = "IW14 Use of unique_together instead of Options.constraints"


class TogetherCheckFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.in_meta = []
        self.issues: list[tuple] = []

        super().__init__()

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        if node.name == "Meta":
            self.in_meta.append(True)

        self.generic_visit(node)

        if node.name == "Meta":
            self.in_meta.pop()

        return node

    def visit_Name(self, node: ast.Name) -> ast.Name:
        if self.in_meta and node.id == "unique_together":
            self.issues.append(
                (node.lineno, node.col_offset, _IW_ERROR_UNIQUE_TOGETHER)
            )
        if self.in_meta and node.id == "index_together":
            self.issues.append(
                (node.lineno, node.col_offset, _IW_ERROR_INDEX_TOGETHER)
            )

        self.generic_visit(node)
        return node



import ast

_IW_TIME_PATCH_CALL_ERROR_MSG = "IW02 unsafe time patch call"


class ForbiddenTimePatchFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []
        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        if isinstance(node.func, ast.Name) and len(node.args) == 1 and node.func.id == "patch":
            patch_target = node.args[0]

            if isinstance(patch_target, ast.Constant) and patch_target.value == "django.utils.timezone.now":
                self.issues.append((node.func.lineno, node.func.col_offset, _IW_TIME_PATCH_CALL_ERROR_MSG))

        if isinstance(node.func, ast.Attribute) and len(node.args) == 1 and node.func.attr == "patch":
            patch_target = node.args[0]

            if isinstance(patch_target, ast.Constant) and patch_target.value == "django.utils.timezone.now":
                self.issues.append((node.func.value.lineno, node.func.value.col_offset, _IW_TIME_PATCH_CALL_ERROR_MSG))

        self.generic_visit(node)
        return node



import sys

from .bulk_batch_update import BulkUpdateFinder
from .datetime_now import DatetimeNowFinder
from .datetime_replace_tzinfo import DatetimeReplaceTzinfoFinder
from .factory_timezone_now import FactoryTimezoneNowFinder
from .freeze_time import ForbiddenTimePatchFinder
from .logging import ErrorLoggerFinder
from .patch_call import ForbiddenPatchCallFinder
from .shared_task import SharedTaskFinder
from .timezone_activate import TimezoneActivateNowFinder
from .foreign_key_check import ForeignKeyCheckFinder
from .constraint_check import ConstraintCheckFinder
from .together_check import TogetherCheckFinder
from .pytest_setup_class import PytestSetupAndTearDownFinder
from .parameterized_order import ParameterizedOrderFinder

if sys.version_info < (3, 8):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


class Plugin(object):
    name = "flake8_iw"
    version = importlib_metadata.version("flake8_iw")
    rules = (
        ForbiddenPatchCallFinder,
        ForbiddenTimePatchFinder,
        ErrorLoggerFinder,
        DatetimeNowFinder,
        DatetimeReplaceTzinfoFinder,
        BulkUpdateFinder,
        SharedTaskFinder,
        TimezoneActivateNowFinder,
        ForeignKeyCheckFinder,
        ConstraintCheckFinder,
        TogetherCheckFinder,
        FactoryTimezoneNowFinder,
        PytestSetupAndTearDownFinder,
        ParameterizedOrderFinder,
    )

    def __init__(self, tree) -> None:
        self.tree = tree

    def run(self):
        for rule_class in self.rules:
            parser = rule_class(self.tree)
            parser.visit(self.tree)

            for lineno, column, msg in parser.issues:
                yield (lineno, column, msg, Plugin)



import ast

_IW_ERROR_TIMEZONE_ACTIVATE_MSG = ("IW08 use of timezone.activate or timezome.deactivate, "
                                   "use with timezone.override instead")


class TimezoneActivateNowFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        # from django.utils import timezone
        # timezone.activate()
        # timezone.deactivate()
        if (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Name)
                and node.func.value.id == "timezone"
                and node.func.attr in ["activate", "deactivate"]
        ):
            self.issues.append(
                (node.func.value.lineno, node.func.value.col_offset, _IW_ERROR_TIMEZONE_ACTIVATE_MSG)
            )

        # from django import utils
        # utils.timezone.activate()
        # utils.timezone.deactivate()
        if (
                isinstance(node.func, ast.Attribute)
                and isinstance(node.func.value, ast.Attribute)
                and isinstance(node.func.value.value, ast.Name)
                and node.func.value.value.id == "utils"
                and node.func.value.attr == "timezone"
                and node.func.attr in ["activate", "deactivate"]
        ):
            self.issues.append(
                (node.func.value.lineno, node.func.value.col_offset, _IW_ERROR_TIMEZONE_ACTIVATE_MSG)
            )

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_DATETIME_NOW_MSG = "IW04 use of datetime.now"


class DatetimeNowFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []

        super().__init__()

    def visit_Call(self, node: ast.Call) -> ast.Call:
        # from datetime import datetime
        # datetime.now()
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == "datetime"
            and node.func.attr == "now"
        ):
            self.issues.append(
                (node.func.value.lineno, node.func.value.col_offset, _IW_ERROR_DATETIME_NOW_MSG)
            )

        # import datetime
        # datetime.datetime.now()
        if (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Attribute)
            and isinstance(node.func.value.value, ast.Name)
            and node.func.value.value.id == "datetime"
            and node.func.attr == "now"
        ):
            self.issues.append(
                (node.func.value.lineno, node.func.value.col_offset, _IW_ERROR_DATETIME_NOW_MSG)
            )

        self.generic_visit(node)
        return node



import ast

_IW_ERROR_MSG = "IW17 parameterized must be topmost decorator"


def visit_decorators(decorator_list: list, issue_list: list):
    for i, decorator in enumerate(decorator_list):
        if isinstance(decorator, ast.Call) and isinstance(decorator.func, ast.Attribute):
            inner = decorator.func.value
            while isinstance(inner, ast.Attribute):
                inner = inner.value

            if isinstance(inner, ast.Name) and inner.id == "parameterized" and i != 0:
                issue_list.append(
                    (decorator.func.lineno, decorator.func.col_offset, _IW_ERROR_MSG)
                )
                break


class ParameterizedOrderFinder(ast.NodeVisitor):
    def __init__(self, _, *args, **kwargs) -> None:
        self.issues: list[tuple] = []
        super().__init__()

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        visit_decorators(node.decorator_list, self.issues)

        self.generic_visit(node)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        visit_decorators(node.decorator_list, self.issues)

        self.generic_visit(node)
        return node
